parse version "_" majorVersion "."
if majorVersion<5 then
do
   say "this program needs at least ooRexx 5.0 to run, aborting"
   exit -1
end

-- make sure we change the current directory to the one this Rexx program resides in
parse source  . . pgm   -- get fully qualified path to this program

-- change current directory such that relatively addressed resources (fxml-file) can be found
call directory filespec('L', pgm)

 -- set directory to contain objects relevant to this application
.environment~my.app=.directory~new


-- import JavaFX classes that we may use more often in different parts of the program
call bsf.import    "javafx.fxml.FXMLLoader",                      "fx.FXMLLoader"
call bsf.import    "javafx.scene.Scene",                          "fx.Scene"

call bsf.import    "javafx.scene.image.Image",                    "fx.Image"
call bsf.import    "javafx.scene.image.ImageView",                "fx.ImageView"

call bsf.import    "javafx.collections.FXCollections",            "fx.FXCollections"
call bsf.loadClass "javafx.util.Callback",                        "fx.Callback"
call bsf.loadClass "javafx.util.Duration",                        "fx.Duration"

call bsf.import    "javafx.beans.property.SimpleStringProperty",  "fx.SimpleStringProperty"
call bsf.import    "javafx.beans.property.SimpleIntegerProperty", "fx.SimpleIntegerProperty"

call bsf.loadClass "javafx.scene.control.ContentDisplay",         "fx.ContentDisplay"
call bsf.loadClass "javafx.scene.input.KeyEvent",                 "fx.KeyEvent"

call bsf.import    "javafx.scene.control.cell.TextFieldTableCell","fx.TextFieldTableCell"
call bsf.import    "javafx.util.converter.DefaultStringConverter","fx.DefaultStringConverter"
call bsf.import    "javafx.stage.Modality",                       "fx.Modality"

--create Rexx object that will control the application
rexxApp=.RxMainApplication~new

--store the Rexx MainApp object in .my.app
.my.app~mainApp=rexxApp

--instantiate the abstract JavaFX class, the abstract "start" method will be served by rexxApp
jRexxApp=BsfCreateRexxProxy(rexxApp, ,"javafx.application.Application")

signal on syntax
--launch the application, invoke "start" and then stay up until the application closes
jRexxApp~launch(jRexxApp~getClass, .nil)
call sysSleep 1
exit

syntax:
   co=condition("object")
   -- display Java exception chain and stack trace of original Java exception
   say ppJavaExceptionChain(co, .true)
   say " done. "~center(100, "-")
   exit -1


--require BSF.CLS
::requires "BSF.CLS"

/* =================================================================================== */
::class RxMainApplication

::attribute primaryStage
::attribute tableViewData
::attribute xmlFilePath

::method init
  expose tableViewData xmlFilePath
  --setup observableArrayList to store tableView data
  tableViewData = .fx.FXCollections~observableArrayList
  .my.app~tableViewData=tableViewData

  -- setup xmlFilePath
  xmlFilePath=.nil

::method start
  expose locale primaryStage xmlFilePath
  use arg primaryStage

  -- bind app.fxml
  appUrl=.bsf~new("java.net.URL", "file:app.fxml")
  rootNode=.fx.FXMLLoader~load(appUrl)

  --setup scene
  scene=.fx.scene~new(rootNode)
  primaryStage~setScene(scene)

  --init appController to setup and register event handlers
  .appController~new
  --show stage
  primaryStage~show

::method importRecordsFromFile
  expose tableViewData
  use arg filePath

  /* Create an instance of the JAXP DocumentBuilderFactory */
  factory=bsf.loadClass("javax.xml.parsers.DocumentBuilderFactory")~newInstance
  /* Set the parser to namespace aware */
  factory~setNamespaceAware(.true)
  /* Ignore whitespaces */
  factory~setIgnoringElementContentWhitespace(.true)
  /* Create the parser from the factory */
  parser=factory~newDocumentBuilder

  /* Create an error handler Rexx object */
  eH = .errorHandler~new
  /* Wrap up the Rexx error handler as a Java object */
  javaEH = BsfCreateRexxProxy(eH, , "org.xml.sax.ErrorHandler")
  /* Set the error handler for this parser */
  parser~setErrorHandler(javaEH)

  /* Parse the file */
  rootNode=parser~parse(filePath)

  /* Load the Java interface class */
  nodeCLS = bsf.loadClass("org.w3c.dom.Node")
  /* Make important constants available by storing them in .local */
  .local~CDATA_SECTION_NODE = nodeCLS~CDATA_SECTION_NODE
  .local~TEXT_NODE = nodeCLS~TEXT_NODE
  .local~ELEMENT_NODE = nodeCLS~ELEMENT_NODE

  /* Define initial variables and arrays */
  countBig = 1
  arrBig = .array~new()
  countSmall = 1
  arrSmall = .array~new()
  /* Node with the first record */
  firstElement = rootNode~getElementsByTagName("record")~item(0)

  /* Call routine "followNode" */
  call followNode rootNode, firstElement, arrBig, arrSmall, countBig, countSmall
  /* Assign the return values of "followNode" to the variables */
  arrSmall = result[2]
  countBig = result[3]
  arrBig = result[4]

  /* Add the last record to the array */
  arrBig~put(arrSmall, countBig)

  /* Final array arrBig -> for each array element create new record and overwrite its
     default data with the element's data, then add it to the tableView */
  do i over arrBig
    tmp =.record~new
    .my.app~mainApp~importNewRecord(tmp, i)
    .my.app~mainApp~tableViewData~add(tmp)
  end

/* Walks the document tree recursively */
::routine followNode
  use arg node, firstElement, arrBig, arrSmall, countBig, countSmall
  /* Call routine "processNode" */
  call processNode node, firstElement, arrBig, arrSmall, countBig, countSmall
  /* Assign the return values of "processNode" to the variables */
  node = result[1]
  arrSmall = result[2]
  countSmall = result[3]
  arrBig = result[4]
  countBig = result[5]
  /* Get NodeList with child nodes */
  if node~hasChildNodes then
    do
      children = node~getChildNodes
      loop i = 0 to children~length-1
        /* Call routine "followNode" */
        call followNode children~item(i), firstElement, arrBig, arrSmall, countBig, countSmall
        /* Assign the return values of "followNode" to the variables */
        countSmall = result[1]
        arrSmall = result[2]
        countBig = result[3]
      end
    end
  /* Return values */
  return countSmall, arrSmall, countBig, arrBig


/* Processes each node */
::routine processNode
  use arg node, firstElement, arrBig, arrSmall, countBig, countSmall
  /* Get the type of the node */
  nodeType = node~getNodeType
  /* If the current node includes text */
  if (nodeType = .text_node | nodeType = .cdata_section_node) then
    do
      /* Add the text to the array */
      arrSmall~put(node~nodeValue, countSmall)
      /* Increase the counter by one */
      countSmall = countSmall + 1
    end
  /* If the current node is an element node, has the name "record" and is not the first "record" */
  if (nodeType = .element_node & node~getNodeName = "record" & \node~isEqualNode(firstElement)) then
    do
      /* Add the current record to the array */
      arrBig~put(arrSmall, countBig)
      /* Increase the counter by one */
      countBig = countBig + 1
      /* Create a new empty array for the next record */
      arrSmall = .array~new()
      /* Reset the counter */
      countSmall = 1
    end

  /* Return values */
  return node, arrSmall, countSmall, arrBig, countBig

::method saveRecordsToFile
  expose tableViewData primaryStage
  use arg filePath

  --setup DocumentBuilder, Document and Parser
  factory=bsf.loadClass("javax.xml.parsers.DocumentBuilderFactory")~newInstance
  factory~setNamespaceAware(.true)
  parser=factory~newDocumentBuilder
  document=parser~newDocument

  --create root xml element and append into document
  root=document~createElement("recordset");
  document~appendChild(root);

  --for each entry in tableViewData, setup elements and assign values, add these to the xml
  do r over tableViewData
    -- create root element for record
    record=document~createElement("record")
    -- create identifier
    identifier=document~createElement("identifier")
    -- add identifier to record
    identifier~appendChild(document~createTextNode(r~nr))
    record~appendChild(identifier)

    -- date
    date=document~createElement("date")
    date~appendChild(document~createTextNode(r~date))
    record~appendChild(date)

    -- recordText
    text = document~createElement("text")
    text~appendChild(document~createTextNode(r~recordText))
    record~appendChild(text)

    -- net
    net = document~createElement("netvalue")
    net~appendChild(document~createTextNode(r~net))
    record~appendChild(net)

    -- vatpct
    vatpct=document~createElement("vatpct")
    vatpct~appendChild(document~createTextNode(r~vatpct))
    record~appendChild(vatpct)

    -- vat (value)
    vat=document~createElement("vatvalue")
    vat~appendChild(document~createTextNode(r~vat))
    record~appendChild(vat)

    -- total
    total=document~createElement("totalvalue")
    total~appendChild(document~createTextNode(r~total))
    record~appendChild(total)

    -- tenant
    tenant=document~createElement("tenant")
    tenant~appendChild(document~createTextNode(r~tenant))
    record~appendChild(tenant)

    -- e1a
    e1a=document~createElement("e1aid")
    e1a~appendChild(document~createTextNode(r~e1a))
    record~appendChild(e1a)

    -- append record node and its content to the root node
    root~appendChild(record)
  end

  --setup TransformerFactory, Transformer, DOMSource and StreamResult objects to write the xml file.
  transformerFactory=bsf.loadClass("javax.xml.transform.TransformerFactory")~newInstance
  transformer=transformerFactory~newTransformer
  outputkeys=bsf.loadClass("javax.xml.transform.OutputKeys")
  transformer~setOutputProperty(outputkeys~INDENT, "yes");
  transformer~setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
  domImpl=bsf.loadClass("org.w3c.dom.DOMImplementation")
  domImpl=document~getImplementation
  doctype=bsf.loadClass("org.w3c.dom.DocumentType")
  doctype = domImpl~createDocumentType("test", .nil, "data.dtd")

  transformer~setOutputProperty(outputkeys~DOCTYPE_SYSTEM, doctype~getSystemId);

  domSource=.bsf~new("javax.xml.transform.dom.DOMSource", document)
  fp=.bsf~new("java.io.File", filepath)
  streamResult=.bsf~new("javax.xml.transform.stream.StreamResult", fp)
  transformer~transform(domSource, streamResult);

  /* Create a window which signals that the export of the XML file was successful */

  /* Create an instance of the class CodeDraw and set the canvas size */
  cd = .bsf~new("codedraw.CodeDraw", 500, 500)

  /* Set the text which is displayed in the top left corner of the CodeDraw window */
  cd~setTitle("BookKeeper")

  /* Get the text formatting options of this CodeDraw window */
  format = cd~getTextFormat()
  /* Define the font size of the drawn text */
  format~setFontSize(40)
  /* Draw the text at the specified coordinate */
  cd~drawText(95, 100, "Export successful!")

  /* Initialize the package "Palette" with "loadClass" */
  color = bsf.loadClass("codedraw.Palette")
  /* Set the drawing color to lime green */
  cd~setColor(color~LIME_GREEN)
  /* Draw a filled circle */
  cd~fillCircle(250, 320, 100)
  /* Set the drawing color to forest green */
  cd~setColor(color~FOREST_GREEN)
  /* Define the thickness of the drawn shape */
  cd~setLineWidth(5)
  /* Draw the outline of the circle */
  cd~drawCircle(250, 320, 110)

  /* Set the drawing color to white */
  cd~setColor(color~WHITE)
  /* Define the thickness of the drawn lines */
  cd~setLineWidth(15)
  /* Draw straight lines between start and end points */
  cd~drawLine(240, 350, 280, 280)
  cd~drawLine(207, 330, 240, 350)

  /* Set the distance in pixel from the top left corner of the screen to the top left corner of the CodeDraw window */
  cd~setWindowPositionX(350)
  cd~setWindowPositionY(70)

  /* Display the drawn shapes on the canvas for 2000 milliseconds (2 seconds),
     then close the CodeDraw window */
  cd~show()
  call sysSleep 2
  cd~dispose(.false)


::method importNewRecord
  use arg record, data
  dataController = .DataController~new(record, data)

::method showRecordEditDialog
  expose redDialog redController primaryStage
  use arg record

  if \var("redDIALOG") then
  do
     -- Load the fxml file and create a new stage for the popup
     recordEditDialogUrl=.bsf~new("java.net.URL", "file:RecordEditDialog.fxml")
     page  = .fx.FXMLLoader~load(recordEditDialogUrl)
     scene = .fx.Scene~new(page)

     dialogStage = .bsf~new("javafx.stage.Stage")
     dialogStage~setTitle("New Record")
     dialogStage~initModality(.fx.Modality~WINDOW_MODAL)

     dialogStage~initOwner(primaryStage)
     dialogStage~setScene(scene)

     -- Set the record into the controller
     redController = .RecordEditDialogController~new
     redController~dialogStage=dialogStage
  end

  redController~setRecord(record)
  redController~okClicked=.false

  -- Show the dialog and wait until the user closes it
  dialogStage~showAndWait
  return redController~OkClicked

::class appController

::attribute id_tableview
::attribute id_column_nr
::attribute id_column_date
::attribute id_column_net
::attribute id_column_vatpct
::attribute id_column_vat
::attribute id_column_total
::attribute id_column_tenant
::attribute id_column_e1a
::attribute id_column_text

::method init
  expose id_vbox id_tableview id_menu_export id_menu_import id_menu_quit  -
         id_column_nr id_column_date id_column_net id_column_vatpct id_column_vat id_column_total -
         id_column_tenant id_column_e1a id_column_text -
         id_button_new id_button_export

  -- fetch the directory containing the JavaFX fx:id JavaFX objects and assign them to attributes
  appDir=.my.app~app.fxml

  .my.app~mainApp~xmlFilePath=.nil
  .my.app~mainApp~primaryStage~title="BookKeeper"

  id_tableview      =appDir~id_tableview
  id_vbox           =appDir~id_vbox

  id_menu_import    =appDir~id_menu_import
  id_menu_export    =appDir~id_menu_export
  id_menu_quit      =appDir~id_menu_quit
  id_menue          =appDir~id_menu

  id_column_nr      =appDir~id_column_nr
  id_column_date    =appDir~id_column_date
  id_column_text    =appDir~id_column_text
  id_column_net     =appDir~id_column_net
  id_column_vatpct  =appDir~id_column_vatpct
  id_column_vat     =appDir~id_column_vat
  id_column_total   =appDir~id_column_total
  id_column_tenant  =appDir~id_column_tenant
  id_column_e1a     =appDir~id_column_e1a

  --add self as the EventHandler (we have the method "handle" implemented)
  rp=BSFCreateRexxProxy(self, ,"javafx.event.EventHandler")

  /* set callbacks for filling the individual TableView cells */
  id_column_date~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("dateProperty"), , "javafx.util.Callback"))
  id_column_nr~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("nrProperty"), ,"javafx.util.Callback"))
  id_column_net~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("netProperty"), ,"javafx.util.Callback"))
  id_column_vatpct~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("vatpctProperty"), ,"javafx.util.Callback"))
  id_column_vat~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("vatProperty"), ,"javafx.util.Callback"))
  id_column_total~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("totalProperty"), ,"javafx.util.Callback"))
  id_column_tenant~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("tenantProperty"), ,"javafx.util.Callback"))
  id_column_e1a~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("e1aProperty"), ,"javafx.util.Callback"))
  id_column_text ~setCellValueFactory(BsfCreateRexxProxy(.PropertyValueFactory~new("recordTextProperty"),  ,"javafx.util.Callback"))

  /* set callback for menu click */
  id_menu_import    ~setOnAction(rp)
  id_menu_export    ~setOnAction(rp)
  id_menu_quit      ~setOnAction(rp)

  /* set callback for button press actions */
  id_button_new     =appdir~id_button_new    ~~setOnAction(rp)
  id_button_export  =appdir~id_button_export ~~setOnAction(rp)

  --fill TableView
  self~setItems

::method setItems
  expose id_tableview
  id_tableview~setItems(.my.app~mainApp~tableViewData)

::method handleNewRecord
    tempRec=.record~new

    okClicked = .my.app~mainApp~showRecordEditDialog(tempRec)
    if okClicked then
        .my.app~mainApp~tableViewData~add(tempRec)

::method handleImport
  expose mainApp

    fileChooser=.bsf~new("javafx.stage.FileChooser")
    -- crate an array of file extensions (only one in this case)
    jarrExtensions=bsf.createJavaArrayOf("java.lang.String", "*.xml")
    --set extensionFilter
    extFilter=.bsf~new("javafx.stage.FileChooser$ExtensionFilter", "XML files (*.xml)", jarrExtensions)
    fileChooser~getExtensionFilters~add(extFilter)
    --set directory to be shown when opening the dialog
    fileChooser~setInitialDirectory(.bsf~new("java.io.File", "."))

    -- Show save file dialog
    file = fileChooser~showOpenDialog(.my.app~mainApp~primaryStage)
    if file<>.nil then
    do
       filePath=file~getPath
       if filePath~right(4)<>".xml" then   -- Make sure it has the correct extension
          filePath||=".xml"
       .my.app~mainApp~importRecordsFromFile( filePath )
    end

::method handleExport
  expose mainApp

  fileChooser=.bsf~new("javafx.stage.FileChooser")
  -- crate an array of file extensions (only one in this case)
  jarrExtensions=bsf.createJavaArrayOf("java.lang.String", "*.xml")
  --set extensionFilter
  extFilter=.bsf~new("javafx.stage.FileChooser$ExtensionFilter", "XML files (*.xml)", jarrExtensions)
  fileChooser~getExtensionFilters~add(extFilter)
  --set directory to be shown when opening the dialog
  fileChooser~setInitialDirectory(.bsf~new("java.io.File", "."))

  -- Show save file dialog
  file = fileChooser~showSaveDialog(.my.app~mainApp~primaryStage)
  if file<>.nil then
  do
     filePath=file~getPath
     if filePath~right(4)<>".xml" then   -- Make sure it has the correct extension
        filePath||=".xml"
     .my.app~mainApp~saveRecordsToFile( filePath )
  end

--implements the interface "javafx.event.EventHandler"
::method handle
  expose id_tableview id_menu_export id_menu_import id_menu_quit -
         id_button_new id_button_export
  use arg event

  source=event~source
  target=event~target
  eventType=event~eventType~toString
  isMouseEvent=(event~objectname~startsWith("javafx.scene.input.MouseEvent@"))

  /* Handle menu and button events. Note: the id values of the context MenuItems are set
     to the same id values defined in the FXML file such that the same actions get carried
     out. */

  --get fx:id value
  tgtId=target~getId
  --fetch observable array list that backs the TableView
  tableViewData=.my.app~tableViewData

  select
     when tgtId="id_menu_export" then
        self~handleExport

     when tgtId="id_menu_import" then
        self~handleImport

     when tgtId="id_menu_quit" then
           do
              self~handleExit
              return
           end

      when tgtId="id_button_new" then
        do
            self~handleNewRecord
        end

      when tgtId="id_button_export" then
        self~handleExport

      when eventType="EDIT_CANCEL" then    -- force a refresh to make sure UI gets syncrhonized with state
           do
              id_tableview~refresh        -- force a refresh of the TableView
              return
           end

     otherwise nop
  end

  if eventType~startsWith("EDIT") then    -- handle edit_cancel event, show information about edit events
  do
     idx=event~getIndex
     if eventType="EDIT_CANCEL" then
        id_tableview~refresh
  end

 /** Closes the JavaFX application. */
::method handleExit
  --unload JavaFX, but let Rexx continue in main thread
  bsf.loadClass("javafx.application.Platform")~exit
/* =================================================================================== */

::class Record

--attributes
::attribute date get
  expose date
  return date~get

::attribute date set
  expose date
  use arg val
  return date~set(val)

::attribute dateProperty get
  expose date
  return date

::attribute nr get
  expose nr
  return nr~get

::attribute nr set
  expose nr
  use arg val
  return nr~set(val)

::attribute nrProperty get
  expose nr
  return nr

::attribute recordText get
  expose recordText
  return recordText~get

::attribute recordText set
  expose recordText
  use arg val
  return recordText~set(val)

::attribute recordTextProperty get
  expose recordText
  return recordText

::attribute net get
  expose net
  return net~get

::attribute net set
  expose net
  use arg val
  return net~set(val)

::attribute netProperty get
  expose net
  return net

::attribute vatpct get
  expose vatpct
  return vatpct~get

::attribute vatpct set
  expose vatpct
  use arg val
  return vatpct~set(val)

::attribute vatpctProperty get
  expose vatpct
  return vatpct

::attribute vat get
  expose vat
  return vat~get

::attribute vat set
  expose vat
  use arg val
  return vat~set(val)

::attribute vatProperty get
  expose vat
  return vat

::attribute total get
  expose total
  return total~get

::attribute total set
  expose total
  use arg val
  return total~set(val)

::attribute totalProperty get
  expose total
  return total

::attribute tenant get
  expose tenant
  return tenant~get

::attribute tenant set
  expose tenant
  use arg val
  return tenant~set(val)

::attribute tenantProperty get
  expose tenant
  return tenant

::attribute e1a get
  expose e1a
  return e1a~get

::attribute e1a set
  expose e1a
  use arg val
  return e1a~set(val)

::attribute e1aProperty get
  expose e1a
  return e1a

::method init
  expose date nr recordText net vatpct vat total tenant e1a

    nr          =.fx.SimpleStringProperty~new("ER 99")
    date        =.fx.SimpleStringProperty~new("2022-06-16")
    recordText  =.fx.SimpleStringProperty~new("A1/Abrg. 06/22")
    net         =.fx.SimpleIntegerProperty~new(20)
    vatpct      =.fx.SimpleIntegerProperty~new(20)
    vat         =.fx.SimpleIntegerProperty~new(4)
    total       =.fx.SimpleIntegerProperty~new(24)
    tenant      =.fx.SimpleStringProperty~new("A1 Telekom")
    e1a         =.fx.SimpleStringProperty~new("9230")

/* =================================================================================== */
::class DataController
::attribute record
::attribute data

::method init
  expose record
  use arg record, data
    --say data[1]

    record~date       =data[2]
    record~nr         =data[1]
    record~net        =data[4]
    record~vatpct     =data[5]
    record~vat        =data[6]
    record~total      =data[7]
    record~tenant     =data[8]
    record~e1a        =data[9]
    record~recordText =data[3]

    --.my.app~mainApp~tableViewData~add(tmp)

::class RecordEditDialogController
--attributes
::attribute okClicked
::attribute dialogStage
::attribute record

::method init
  expose idField dateField recordTextField netField vatpctField vatField totalField tenantField e1aField -
         btnOK btnCancel okClicked

  okClicked=.false

  -- fetch the directory containing the JavaFX fx:id JavaFX objects and assign them to attributes
  redDir            =.my.app~RecordEditDialog.fxml
  dateField         =redDir~dateField
  idField           =redDir~idField
  netField          =redDir~netField
  vatpctField       =redDir~vatpctField
  vatField          =redDir~vatField
  totalField        =redDir~totalField
  tenantField       =redDir~tenantField
  e1aField          =redDir~e1aField
  recordTextField   =redDir~recordTextField

  btnOK          =redDir~btnOK
  btnCancel      =redDir~btnCancel

   -- add self as the EventHandler (method "handle" implemented)
  rp=BSFCreateRexxProxy(self, ,"javafx.event.EventHandler")
   -- add us as the event handler to the buttons, the event handler (method "handle") will be able to get at the pressed button
  btnOK    ~setOnAction(rp)
  btnCancel~setOnAction(rp)


::method handle
  expose btnOK btnCancel
  use arg event

  tgtObjectName=event~getTarget~objectname
  select
     when tgtObjectName=btnOK~objectName     then self~handleOK
     when tgtObjectName=btnCancel~objectName then self~handleCancel
     otherwise .error~say("UNKNOWN target:" pp(event~getTarget) pp(event~getTarget~tostring))
  end


::method setRecord
  expose idField dateField recordTextField netField vatpctField vatField totalField tenantField e1aField record
  use arg record

  dateField~setText(record~date)
  idField~setText(record~nr)
  netField~setText(record~net)
  vatpctField~setText(record~vatpct)
  vatField~setText(record~vat)
  totalField~setText(record~total)
  tenantField~setText(record~tenant)
  e1aField~setText(record~e1a)
  recordTextField~setText(record~recordText)

::method handleCancel
  expose dialogStage

  dialogStage~hide

::method handleOK
  expose idField dateField recordTextField netField vatpctField vatField totalField tenantField e1aField -
         record okClicked dialogStage

  record~date       =dateField~getText
  record~nr         =idField~getText
  record~net        =netField~getText
  record~vatpct     =vatpctField~getText
  record~vat        =vatField~getText
  record~total      =totalField~getText
  record~tenant     =tenantField~getText
  record~e1a        =e1aField~getText
  record~recordText =recordTextField~getText

  okClicked = .true
  dialogStage~hide

::class PropertyValueFactory
::method init
  expose columnName
  use strict arg columnName

::method call
  expose columnName
  use arg o

  if o~getValue~isNil then return .nil
  aRexxObj=BsfRexxProxy(o~getValue) -- unbox Rexx object
  return aRexxObj~send(columnName)  -- get attribute's value from the Rexx object


/* =================================================================================== */
::class "RexxCellFactory" -- Rexx class implementing "javafx.util.Callback"'s "call" method

::method init
  expose columnName
  use strict arg columnName

::method call           -- implements "javafx.util.Callback.call(Object o)": in this case an instance of a TableCell (or a subclass of it) must be returned!
  expose columnName     -- the name of the column/attribute we have to serve

  jCell=.fx.TextFieldTableCell~new
  if columnName<>"nr" then
     jCell~setConverter(.fx.DefaultStringConverter~new)  -- assign string converter to use

  jCell~setTooltip(.my.app~cell.tooltip)     -- assign tooltip

  return jCell         -- return the Java cell

/* =================================================================================== */
--find and return cell object from an event object.
::routine findCellObject
   use arg event
      -- get the node of the event
   signal on syntax     -- in case event does not have the pickResult method
   node=event~pickResult~getIntersectedNode
   do while \node~isNil    -- find the Cell node going up the hierarchy
      if node~objectname~pos("Cell")>0 then  -- we found the Cell node
         return node
      node=node~getParent  -- get next parent node
   end
   return .nil

syntax:
   if .my.app~showDebugOutput=.true then
      say .line .dateTime~new .context~name"(event): SYNTAX condition (pickResult-method not present)"
   return .nil

/* =================================================================================== */
--find and return the cell's index value, if any.
::routine getIndex
   use arg event
      -- get the node of the event
   cell=findCellObject(event)
   if cell~isNil then return -1
   return cell~getIndex

/* SAX Error handler */
::class ErrorHandler
::method unknown
  use arg methName, argArray
  exception = argArray[1]
  .error~say(methName": " -
            "line = "exception~getLineNumber", col = "exception~getColumnNumber": " -
            pp(exception~getMessage))
